!!!
! Parallel Hybrid PingPong with threads support level of type Funneled
! Computation of the sustained network bandwidth between two nodes,
! with nbMPIProcs/2 MPI process per node and as many parallel streams of communication
! Execution on 2 nodes.
!!!

program PingPongParFun
  ...
  ...

  implicit none

  integer, parameter :: sizeOMPComm = 8
  integer, dimension(sizeOMPComm) :: OMPComm = (/ 1, 2, 3, 4, 8, 16, 32, 64 /)
  integer :: myMPIRank, nbMPIProcs, myOMPRank, nbOMPThds, nbOMPComm
  integer :: sendRank, recvRank, tagPing, tagPong
  integer :: ierr, i, it, level_mpi_provided
  integer, dimension(3), parameter :: tailleBufMo = (/ 1, 10, 100 /)
  real(kind=kind(1.d0)), dimension(:), allocatable :: pingSendBuf, pingRecvBuf, pongRecvBuf

  ! MPI Initialization
  ...
  ! Set myMPIRank & nbMPIProcs
  ...
  ...
  if (level_mpi_provided < MPI_THREAD_FUNNELED) then
     if (myMPIRank == 0) print *, 'Error: level_mpi_provided < MPI_THREAD_FUNNELED'
     call MPI_Abort(MPI_COMM_WORLD, 1, ierr)
  end if

  ! OpenMP // region
  ...
  ! Rank and number of OpenMP threads per MPI process
  myOMPRank =
  nbOMPThds = 

  ! Output information of the run
  if (...) then
     ...
     print *
     print *, "Execution sur ", nbMPIProcs, " processus MPI constitues de ", &
              nbOMPThds, " threads OpenMP"
     print *
     ...
  endif

  ! Hybrid synchronization barrier
  call barrierMPIOMP()

  ! Loop on the message sizes (1 MB, 10 MB and 100 MB)
  do i=1,size(tailleBufMo,1)
     call pingPongPar(tailleBufMo(i), nbOMPThds, myOMPRank)
  enddo

  ! End of OpenMP // region
  ...

  ! Finalization of MPI environment
  call MPI_FINALIZE(ierr)

  contains

    subroutine pingPongPar(tailleMo, nbOMPThds, myOMPRank)
      integer, intent(in) :: tailleMo, nbOMPThds, myOMPRank
      ! Local variables
      integer :: iter, nbElBuf, nbElBufLoc, nbIter
      real(kind=kind(1.d0)) :: startTime, endTime
      integer, dimension(MPI_STATUS_SIZE) :: statut

      ! Number of elements of buffers
      nbElBufLoc = 1e6 * tailleMo / (8 * nbOMPThds)
      nbElBuf = nbElBufLoc * nbOMPThds
      ! Number of iterations for a total data transfert of 1 GB pingPong
      nbIter = 1000 / tailleMo

      ...
      ! Allocations of send and receive buffers
      if (...) then
         ! 1st half are sending and receiving
         allocate(pingSendBuf(nbElBuf),pongRecvBuf(nbElBuf))
      else
         ! 2nd half are receiving and sending using the same buffer
         allocate(pingRecvBuf(nbElBuf))
      endif
      ! Who to send (sendRank), who to receive (recvRank) ?
      sendRank = ...
      recvRank = ...
      tagping = 1
      tagpong = 2
      ...

      ! // Initialization of pingSendBuf, each thread initializes
      ! a slice of buffer pingSendBuf
      if (...) &
           pingSendBuf(1 + (myOMPRank * nbElBufLoc):(myOMPRank + 1) * nbElBufLoc) = myOMPRank
      ...

      ! Communications between MPI processes
      ...
      call MPI_BARRIER(MPI_COMM_WORLD, ierr)
      ! Timer for the // PingPong on node 0
      if (myMPIRank == 0) startTime = MPI_WTIME()

      ! Loop to get the equivalent of 1 GB data transfert per pingpong
      do iter=1,nbIter
         if (...) then
            ! Ping : Send pingSendBuf to sendRank
            ...
            ! Pong : Receive from sendRank in pongRecvBuf
            ...
         else
            ! Ping : Receive from sendRank in pingRecvBuf
            ...
            ! Pong : Send pingRecvBuf to sendRank
            ...
         endif
         ! Waiting for Pong completion
         ...
      enddo

      ! End of timing and output of sustained bandwidth
      if (myMPIRank == 0) then
         endTime = MPI_WTIME()
         write(*,'("Taille msg =",I5," Mo,  tps elapsed =",1pe10.3," debit=",1pe10.3," en Mo/s")') &
              tailleMo, endTime - startTime, (nbMPIProcs / 2) * nbIter * 2 * tailleMo / (endTime - startTime)
      endif

      ! Deallocations of send and receive buffers
      if (...) then
         deallocate(pingSendBuf,pongRecvBuf)
      else
         deallocate(pingRecvBuf)
      endif
      ...
      ! Hybrid synchronization barrier
      call barrierMPIOMP()
    end subroutine pingPongPar

    subroutine barrierMPIOMP()
      !$OMP BARRIER
      !$OMP MASTER
      call MPI_BARRIER(MPI_COMM_WORLD, ierr)
      !$OMP END MASTER
      !$OMP BARRIER
    end subroutine barrierMPIOMP

end program PingPongParFun
