!!!
! Parallel Hybrid PingPong with threads support level of type Multiple
! Computation of the sustained network bandwidth between two nodes,
! with one MPI process per node and nbOMPComm parallel streams of communication
! Execution on 2 nodes.
!!!

program PingPongParMult
  ...
  ...

  implicit none

  integer, parameter :: sizeOMPComm = 8
  integer, dimension(sizeOMPComm) :: OMPComm = (/ 1, 2, 3, 4, 8, 16, 20, 40 /)
  integer :: myMPIRank, nbMPIProcs, myOMPRank, nbOMPThds, nbOMPComm
  integer :: ierr, i, it, level_mpi_provided
  integer, dimension(3), parameter :: tailleBufMo = (/ 1, 10, 100 /)

  ! MPI Initialization
  ...
  ! Get myMPIRank & nbMPIProcs
  ...
  ...
  if (level_mpi_provided < MPI_THREAD_MULTIPLE) then
     if (myMPIRank == 0) print *, 'Error: level_mpi_provided < MPI_THREAD_MULTIPLE'
     call MPI_Abort(MPI_COMM_WORLD, 1, ierr)
  end if

  ! OpenMP // region
  ...
  ! Rank and number of OpenMP threads per MPI process
  myOMPRank = ...
  myOMPThds = ...

  ! Output information of the run
  if (myMPIRank == 0) then
     ...
     print *
     print *, "Execution sur ", nbMPIProcs, " processus MPI constitues de ", &
              nbOMPThds, " threads OpenMP"
     print *
     ...
  endif

  ! Loop on the number of tests (1 test for each # of data streams in //)
  do it=1,sizeOMPComm
     if(OMPComm(it) > nbOMPThds) exit
     ! Number of active threads involved in communication
     ...
     nbOMPComm = OMPComm(it)
     if (myMPIRank == 0) print *, "Nombre de threads de communication : ", nbOMPComm
     ...
     ! Hybrid synchronization barrier
     call barrierMPIOMP()
     ! Loop on the message sizes (1 MB, 10 MB and 100 MB)
     do i=1,size(tailleBufMo,1)
        call PingPongPar(nbOMPComm, tailleBufMo(i), myMPIRank, myOMPRank)
     enddo
     if (myMPIRank == 0 .and. myOMPRank == 0) print *
     call barrierMPIOMP()
  enddo

  ! End of OpenMP // region
  ...

  ! Finalization of MPI environment
  ...

  contains

    subroutine pingPongPar(nbOMPComm, tailleMo, myMPIRank, myOMPRank)
      integer, intent(in) :: nbOMPComm, tailleMo, myMPIRank, myOMPRank
      ! Local variables
      integer :: ierr, sendRank, recvRank, tagPing, tagPong, iter, nbElBuf, nbIter
      real(kind=kind(1.d0)), dimension(:), allocatable :: pingSendBuf, pingRecvBuf, pongRecvBuf
      real(kind=kind(1.d0)) :: startTime, endTime
      integer, dimension(MPI_STATUS_SIZE) :: statut

      ! Number of elements in the buffers
      nbElBuf = (1e6 * tailleMo) / 8
      ! Number of iterations for a total data transfert of 1 GB pingPong
      nbIter = 1000 / tailleMo

      ! For communication threads
      if (...) then
         ! Allocations of send and receive buffers
         if (...) then
            allocate(pingSendBuf(nbElBuf),pongRecvBuf(nbElBuf))
         else
            allocate(pingRecvBuf(nbElBuf))
         endif
         ! Who to send (sendRank), who to receive (recvRank) ?
         sendRank = ...
         recvRank = ...
         tagPing = myOMPRank
         tagPong = myOMPRank

         ! Initialization of the send buffer pingSendBuf
         if (...) then
            pingSendBuf(1:nbElBuf) = myOMPRank
         endif
       endif
       ! Hybrid synchronization barrier
       call barrierMPIOMP()

      ! Timer for the // PingPong on node 0
      !$OMP MASTER
      if (...) startTime = MPI_WTIME()
      !$OMP END MASTER

      ! Loop to get the equivalent of 1 GB data transfert per pingpong
      do iter=1,nbIter
         if (...) then
            ! Ping : Send pingSendBuf to sendRank
            if (...) &
               ...
            ! Pong : Receive of sendRank in pongRecvBuf
            if (...) &
               ...
         else
            ! Ping : Receive from sendRank in pingRecvBuf
            if (...) &
               ...
            ! Pong : Send of pingRecvBuf to sendRank
            if (...) &
               ...
         endif
         ! Waiting for Pong completion
         ...
      enddo

      ! End of timing and output of sustained bandwidth
      !$OMP MASTER
      if (...) then
         endTime = MPI_WTIME()
         write(*,'("Taille msg =",I5," Mo,  tps elapsed =",1pe10.3," debit=",1pe10.3," en Mo/s")') &
              tailleMo, endTime - startTime, nbOMPComm * nbIter * 2.0 * tailleMo / (endTime - startTime)
      endif
      !$OMP END MASTER

      ! Deallocations of send and receive buffers
      if (...) then
         if (...) then
            deallocate(pingSendBuf,pongRecvBuf)
         else
            deallocate(pingRecvBuf)
         endif
      endif
      call barrierMPIOMP()
    end subroutine PingPongPar

    subroutine barrierMPIOMP()
      integer :: ierr
      !$OMP BARRIER
      !$OMP MASTER
      call MPI_BARRIER(MPI_COMM_WORLD, ierr)
      !$OMP END MASTER
      !$OMP BARRIER
    end subroutine barrierMPIOMP

end program PingPongParMult
