program barrier_hybride
  use MPI
  use OMP_LIB
  implicit none
  
  integer :: myMPIRank, myOMPRank
  integer :: code, level_mpi_provided
  
  ! Initialisation MPI/OMP
  call MPI_INIT_THREAD(MPI_THREAD_FUNNELED,level_mpi_provided,code)
  ! Recuperation myMPIRank
  call MPI_COMM_RANK(MPI_COMM_WORLD,myMPIRank,code)
  !$OMP PARALLEL private(myOMPRank)
  myOMPRank = OMP_GET_THREAD_NUM()
  
  ! Desynchronisation des threads
  call desync(myMPIRank,myOMPRank)
  write(*,'(" Avant barriere, processus MPI : ",I3," thread OMP : ",I3)') myMPIRank, myOMPRank
  ! Barriere hybride a coder
  call barrierMPIOMP()
  write(*,'(" Apres barriere, processus MPI : ",I3," thread OMP : ",I3)') myMPIRank, myOMPRank
  
  !$OMP END PARALLEL
  ! Fermeture environnement MPI
  call MPI_FINALIZE(code)

  contains

    ! Barriere hybride a coder
    subroutine barrierMPIOMP()
      !$OMP BARRIER
      !$OMP MASTER
      call MPI_BARRIER(MPI_COMM_WORLD,code)
      !$OMP END MASTER
      !$OMP BARRIER
      call sleep(1)
    end subroutine barrierMPIOMP

    ! Desynchronisation des threads, a ne pas toucher
    subroutine desync(myMPIRank,myOMPRank)
      integer, intent(in) :: myMPIRank, myOMPRank
      if (myMPIRank==1 .and. myOMPrank==1) call sleep(1)
    end subroutine desync

end program barrier_hybride
